diff mbox

C++ PATCH for c++/54506 (wrong implicit move)

Message ID 504DF41D.2080203@redhat.com
State New
Headers show

Commit Message

Jason Merrill Sept. 10, 2012, 2:07 p.m. UTC
This area of the standard is in flux, but what we were doing was 
definitely wrong.  The proposed resolution for issue 1402 says that if a 
move constructor would call a non-trivial non-move constructor for a 
subobject, it is not implicitly declared.  We might end up dropping that 
provision entirely, but in any case we should allow moving via template 
constructor as well as non-template.

Tested x86_64-pc-linux-gnu, applying to trunk and 4.7 (since it only 
affects C++11 mode).
diff mbox

Patch

commit 37c8977bb82c984645795a9992fe6658841e2d35
Author: Jason Merrill <jason@redhat.com>
Date:   Mon Sep 10 09:22:37 2012 -0400

    	PR c++/54506
    	* decl.c (move_signature_fn_p): Split out from move_fn_p.
    	* method.c (process_subob_fn): Use it.
    	* cp-tree.h: Declare it.

diff --git a/gcc/cp/cp-tree.h b/gcc/cp/cp-tree.h
index 3e0fc3f..3c55ba4 100644
--- a/gcc/cp/cp-tree.h
+++ b/gcc/cp/cp-tree.h
@@ -5066,6 +5066,7 @@  extern tree build_ptrmem_type			(tree, tree);
 extern tree build_this_parm			(tree, cp_cv_quals);
 extern int copy_fn_p				(const_tree);
 extern bool move_fn_p                           (const_tree);
+extern bool move_signature_fn_p                 (const_tree);
 extern tree get_scope_of_declarator		(const cp_declarator *);
 extern void grok_special_member_properties	(tree);
 extern int grok_ctor_properties			(const_tree, const_tree);
diff --git a/gcc/cp/method.c b/gcc/cp/method.c
index c21ae15..a42ed60 100644
--- a/gcc/cp/method.c
+++ b/gcc/cp/method.c
@@ -947,9 +947,10 @@  process_subob_fn (tree fn, bool move_p, tree *spec_p, bool *trivial_p,
 	}
     }
 
-  /* Core 1402: A non-trivial copy op suppresses the implicit
+  /* Core 1402: A non-trivial non-move ctor suppresses the implicit
      declaration of the move ctor/op=.  */
-  if (no_implicit_p && move_p && !move_fn_p (fn) && !trivial_fn_p (fn))
+  if (no_implicit_p && move_p && !move_signature_fn_p (fn)
+      && !trivial_fn_p (fn))
     *no_implicit_p = true;
 
   if (constexpr_p && !DECL_DECLARED_CONSTEXPR_P (fn))
diff --git a/gcc/testsuite/g++.dg/cpp0x/implicit14.C b/gcc/testsuite/g++.dg/cpp0x/implicit14.C
new file mode 100644
index 0000000..8a56244
--- /dev/null
+++ b/gcc/testsuite/g++.dg/cpp0x/implicit14.C
@@ -0,0 +1,26 @@ 
+// PR c++/54506
+// { dg-do compile { target c++11 } }
+
+template <class T>
+struct A
+{
+  A() {}
+
+  A(A const volatile &&) = delete;
+  A &operator =(A const volatile &&) = delete;
+
+  template <class U> A(A<U> &&) {}
+  template <class U> A &operator =(A<U> &&) { return *this; }
+};
+
+struct B
+{
+  A<int> a;
+  B() = default;
+};
+
+int main()
+{
+  B b = B();
+  b = B();
+}
diff --git a/gcc/cp/decl.c b/gcc/cp/decl.c
index 7655f78..e34092d 100644
--- a/gcc/cp/decl.c
+++ b/gcc/cp/decl.c
@@ -10859,10 +10859,6 @@  copy_fn_p (const_tree d)
 bool
 move_fn_p (const_tree d)
 {
-  tree args;
-  tree arg_type;
-  bool result = false;
-
   gcc_assert (DECL_FUNCTION_MEMBER_P (d));
 
   if (cxx_dialect == cxx98)
@@ -10872,12 +10868,29 @@  move_fn_p (const_tree d)
   if (TREE_CODE (d) == TEMPLATE_DECL
       || (DECL_TEMPLATE_INFO (d)
          && DECL_MEMBER_TEMPLATE_P (DECL_TI_TEMPLATE (d))))
     /* Instantiations of template member functions are never copy
        functions.  Note that member functions of templated classes are
        represented as template functions internally, and we must
-       accept those as copy functions.  */
+       accept those as move functions.  */
     return 0;
 
+  return move_signature_fn_p (d);
+}
+
+/* D is a constructor or overloaded `operator='.
+
+   Then, this function returns true when D has the same signature as a move
+   constructor or move assignment operator (because either it is such a
+   ctor/op= or it is a template specialization with the same signature),
+   false otherwise.  */
+
+bool
+move_signature_fn_p (const_tree d)
+{
+  tree args;
+  tree arg_type;
+  bool result = false;
+
   args = FUNCTION_FIRST_USER_PARMTYPE (d);
   if (!args)
     return 0;
@@ -10872,12 +10868,29 @@  move_fn_p (const_tree d)
       || (DECL_TEMPLATE_INFO (d)
          && DECL_MEMBER_TEMPLATE_P (DECL_TI_TEMPLATE (d))))
-    /* Instantiations of template member functions are never copy
+    /* Instantiations of template member functions are never move
        functions.  Note that member functions of templated classes are
        represented as template functions internally, and we must